@@ -1,9 +1,13 @@
 from __future__ import annotations
 
 import atexit
+import contextlib
 import datetime
+import json
 import logging
+import os
 import pickle
+import random
 import sys
 import threading
 import warnings
@@ -22,7 +26,9 @@ from typing import (
 import attr
 import bitmath
 
-from .determ_hash import HASH_BYTES, determ_hash, hashable
+from charmonium.determ_hash import determ_hash
+from charmonium.freeze import freeze
+
 from .index import Index, IndexKeyType
 from .obj_store import DirObjStore, ObjStore
 from .pickler import Pickler
@@ -34,45 +40,75 @@ from .util import (
     FuncReturn,
     Future,
     GetAttr,
-    ellipsize,
     identity,
     none_tuple,
 )
 
-__version__ = "1.2.1"
+__version__ = "1.2.4"
 
 BYTE_ORDER: str = "big"
 
 
 def memoize(
-    **kwargs: Any,
+    *, disable: Union[str, bool] = None, **kwargs: Any,
 ) -> Callable[[Callable[FuncParams, FuncReturn]], Memoized[FuncParams, FuncReturn]]:
     """See :py:class:`charmonium.cache.Memoized`."""
 
-    def actual_memoize(
-        func: Callable[FuncParams, FuncReturn]
-    ) -> Memoized[FuncParams, FuncReturn]:
-        # pyright doesn't know attrs __init__, hence type ignore
-        return Memoized[FuncParams, FuncReturn](func, **kwargs)  # type: ignore
+    real_disable = (
+        disable
+        if isinstance(disable, bool)
+        else bool(os.environ.get("CHARMONIUM_CACHE_DISABLE", ""))
+    )
 
-    # I believe pyright suffers from this fixed mypy bug: https://github.com/python/mypy/issues/1323
-    # Therefore, I have to circumvent the type system.
-    # However, somehow `cast` isn't sufficient.
-    # Therefore, I need type ignore. I don't like it any more than you.
-    # return cast(Memoized[FuncParams, FuncReturn], actual_memoize)
-    return actual_memoize  # type: ignore
+    if real_disable:
+        return lambda x: x
+    else:
+
+        def actual_memoize(
+            func: Callable[FuncParams, FuncReturn]
+        ) -> Memoized[FuncParams, FuncReturn]:
+            # pyright doesn't know attrs __init__, hence type ignore
+            return Memoized[FuncParams, FuncReturn](func, **kwargs)  # type: ignore
+
+        # I believe pyright suffers from this fixed mypy bug: https://github.com/python/mypy/issues/1323
+        # Therefore, I have to circumvent the type system.
+        # However, somehow `cast` isn't sufficient.
+        # Therefore, I need type ignore. I don't like it any more than you.
+        # return cast(Memoized[FuncParams, FuncReturn], actual_memoize)
+        return actual_memoize  # type: ignore
 
 
 DEFAULT_LOCK_PATH = ".cache/.lock"
 DEFAULT_OBJ_STORE_PATH = ".cache"
 
 
-def key2str(key: Tuple[Any, ...]) -> str:
-    system_key_rest = f" {key[0][1:]}" if len(key[0]) > 1 else ""
-    system_key = f"v{key[0][0]}{system_key_rest}"
-    function = f"{ellipsize(key[1], 20)} ver {determ_hash(key[2]):0{2*HASH_BYTES}x}"
-    args = f"{determ_hash(key[3]):0{2*HASH_BYTES}x} ver {determ_hash(key[4]):0{2*HASH_BYTES}x}"
-    return f"(system={system_key}, function={function}, args={args})"
+ops_logger = logging.getLogger("charmonium.cache.ops")
+# all operations are logged as DEBUG
+perf_logger = logging.getLogger("charmonium.cache.perf")
+# all perf data is logged as DEBUG
+
+perf_logger_file = os.environ.get("CHARMONIUM_CACHE_PERF_LOG")
+if perf_logger_file:
+    perf_logger.setLevel(logging.DEBUG)
+    perf_logger.addHandler(logging.FileHandler(perf_logger_file))
+    perf_logger.propagate = False
+
+
+@contextlib.contextmanager
+def perf_ctx(event: str, **kwargs: Any) -> None:
+    if perf_logger.isEnabledFor(logging.DEBUG):
+        start = datetime.datetime.now()
+    yield
+    if perf_logger.isEnabledFor(logging.DEBUG):
+        perf_logger.debug(
+            json.dumps(
+                {
+                    "event": event,
+                    "duration": (datetime.datetime.now() - start).total_seconds(),
+                    **kwargs,
+                }
+            )
+        )
 
 
 # pyright thinks attrs has ambiguous overload
@@ -97,21 +133,12 @@ class MemoizedGroup:
     time_cost: dict[str, datetime.timedelta]
     time_saved: dict[str, datetime.timedelta]
     temporary: bool
-    logger: logging.Logger
-
-    @classmethod
-    def __determ_hash__(cls) -> Any:
-        # I don't want to put anything about the actual state of Memoized here.
-        # Because then if a Memoized function,f, called a Memoized function, g, then f will think g changed when only the memoization state chaned.
-        # If the memoization is completely transparent, then the memoization state does not affect the behavior of the program/
-        return ()
 
     def __getstate__(self) -> Any:
         return {
             slot: getattr(self, slot)
             for slot in self.__slots__
-            if slot
-            not in {"__weakref__", "_index", "_memory_lock", "_version", "logger"}
+            if slot not in {"__weakref__", "_index", "_memory_lock", "_version"}
         }
 
     def __setstate__(self, state: Mapping[str, Any]) -> Any:
@@ -129,8 +156,6 @@ class MemoizedGroup:
         )
         self._version = 0
         self._memory_lock = threading.RLock()
-        self.logger = logging.getLogger("charmonium.cache")
-        self.logger.setLevel(logging.DEBUG)
         self._index_read()
 
     def __init__(
@@ -200,48 +225,57 @@ class MemoizedGroup:
             key, entry = item
             if entry.obj_store:
                 obj_key = determ_hash(key)
-                self.logger.getChild("invalidate").debug(
-                    "Invalidating key=%s obj_key=0x%x", key2str(key), obj_key
-                )
                 del self._obj_store[obj_key]
+            else:
+                obj_key = None
             self._replacement_policy.invalidate(key, entry)
+        if ops_logger.isEnabledFor(logging.DEBUG):
+            ops_logger.debug(
+                json.dumps(
+                    {"event": "cascading_delete", "key": key, "obj_key": obj_key,}
+                )
+            )
 
     def _index_read(self) -> None:
         with self._memory_lock, self._index_lock.reader:
             self._index_read_nolock()
 
     def _index_read_nolock(self) -> None:
-        if self._index_key in self._obj_store:
-            other_version, other_index, other_rp, other_tc, other_ts = cast(
-                Tuple[
-                    int,
-                    Index[Any, Entry],
-                    ReplacementPolicy,
-                    DefaultDict[str, datetime.timedelta],
-                    DefaultDict[str, datetime.timedelta],
-                ],
-                self._pickler.loads(self._obj_store[self._index_key]),
-            )
-            if other_version > self._version:
-                self.logger.getChild("index_read").debug(
-                    "Reading index, my_version=%d, their_verison=%d",
-                    self._version,
-                    other_version,
+        current_version = self._version
+        with perf_ctx("index_read"):
+            if self._index_key in self._obj_store:
+                other_version, other_index, other_rp, other_tc, other_ts = cast(
+                    Tuple[
+                        int,
+                        Index[Any, Entry],
+                        ReplacementPolicy,
+                        DefaultDict[str, datetime.timedelta],
+                        DefaultDict[str, datetime.timedelta],
+                    ],
+                    self._pickler.loads(self._obj_store[self._index_key]),
                 )
-                self._version = other_version
-                self._index.update(other_index)
-                self._replacement_policy.update(other_rp)
-                self.time_cost = other_tc
-                self.time_saved = other_ts
+                if other_version > current_version:
+                    self._version = other_version
+                    self._index.update(other_index)
+                    self._replacement_policy.update(other_rp)
+                    self.time_cost = other_tc
+                    self.time_saved = other_ts
+        if ops_logger.isEnabledFor(logging.DEBUG):
+            ops_logger.debug(
+                json.dumps(
+                    {
+                        "event": "index_read",
+                        "old_version": current_version,
+                        "self._version": self._version,
+                    }
+                )
+            )
 
     def _index_write(self) -> None:
-        with self._memory_lock, self._index_lock.writer:
+        with perf_ctx("index_write"), self._memory_lock, self._index_lock.writer:
             self._index_read_nolock()
             self._evict()
             self._version += 1
-            self.logger.getChild("_index_write").debug(
-                "Writing index version=%d", self._version
-            )
             self._obj_store[self._index_key] = self._pickler.dumps(
                 (
                     self._version,
@@ -251,6 +285,10 @@ class MemoizedGroup:
                     self.time_saved,
                 )
             )
+        if ops_logger.isEnabledFor(logging.DEBUG):
+            ops_logger.debug(
+                json.dumps({"event": "index_write", "self._version": self._version,})
+            )
 
     def _system_state(self) -> Any:
         """Functions are deterministic with (global state, function-specific state, args key, args version).
@@ -272,19 +310,27 @@ class MemoizedGroup:
 
             while total_size > self._size:
                 key, entry = self._replacement_policy.evict()
-                obj_key = determ_hash(key)
                 if entry.obj_store:
+                    obj_key = determ_hash(key)
                     assert (
                         obj_key in self._obj_store
                     ), "Replacement policy tried to evict something that wasn't there"
-                    self.logger.getChild("_evict").debug(
-                        "Evicting key=%s obj_key=0x%x size=%r",
-                        key2str(key),
-                        obj_key,
-                        entry.data_size.format("{value:.1f}{unit}"),
-                    )
                     del self._obj_store[obj_key]
+                else:
+                    obj_key = None
                 total_size -= entry.data_size
+                if ops_logger.isEnabledFor(logging.DEBUG):
+                    ops_logger.debug(
+                        json.dumps(
+                            {
+                                "event": "evict",
+                                "key": key,
+                                "obj_key": obj_key,
+                                "entry.data_size": entry.data_size.bytes,
+                                "new_total_size": (total_size - entry.data_size).bytes,
+                            }
+                        )
+                    )
                 del self._index[key]
 
     def remove_orphans(self) -> None:
@@ -308,9 +354,10 @@ class MemoizedGroup:
             }
             for obj_key in self._obj_store:
                 if obj_key not in found_obj_keys:
-                    self.logger.getChild("remove_orphans").debug(
-                        "Removing obj_key=0x%x", obj_key
-                    )
+                    if ops_logger.isEnabledFor(logging.DEBUG):
+                        ops_logger.debug(
+                            json.dumps({"event": "remove_orphan", "obj_key": obj_key,})
+                        )
                     del self._obj_store[obj_key]
 
 
@@ -392,16 +439,16 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         with self.group._memory_lock:  # pylint: disable=protected-access
             tc = self.group.time_cost[self.name]
             ts = self.group.time_saved[self.name]
-            print(
-                "Caching %r: cost %.1fs, saved %.1fs, net saved %.1fs"
-                % (
-                    self.name,
-                    tc.total_seconds(),
-                    ts.total_seconds(),
-                    (ts - tc).total_seconds(),
-                ),
-                file=sys.stderr,
-            )
+        print(
+            "Caching %r: cost %.1fs, saved %.1fs, net saved %.1fs"
+            % (
+                self.name,
+                tc.total_seconds(),
+                ts.total_seconds(),
+                (ts - tc).total_seconds(),
+            ),
+            file=sys.stderr,
+        )
 
     @property
     def _pickler(self) -> Pickler:
@@ -432,9 +479,6 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         """
         return (
             self._func,
-            self._pickler,
-            # Group is a "friend class", so pylint disable
-            self.group._obj_store,  # pylint: disable=protected-access
             GetAttr[Callable[[], Any]]()(
                 self._func, "__version__", lambda: None, check_callable=True
             )(),
@@ -469,7 +513,11 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         return f"memoized {self.name}"
 
     def _recompute(
-        self, obj_key: int, *args: FuncParams.args, **kwargs: FuncParams.kwargs
+        self,
+        call_id: int,
+        obj_key: int,
+        *args: FuncParams.args,
+        **kwargs: FuncParams.kwargs,
     ) -> tuple[Entry, FuncReturn]:
 
         start = datetime.datetime.now()
@@ -483,9 +531,10 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             data_size = bitmath.Byte(len(value_ser))
             # Group is a "friend class", hence pylint disable
 
-            self.group._obj_store[  # pylint: disable=protected-access
-                obj_key
-            ] = value_ser
+            with perf_ctx("obj_store", call_id=call_id):
+                self.group._obj_store[  # pylint: disable=protected-access
+                    obj_key
+                ] = value_ser
         else:
             stored_value = value
             data_size = bitmath.Byte(0)
@@ -494,7 +543,26 @@ class Memoized(Generic[FuncParams, FuncReturn]):
 
         # TODO: cache stdout?
 
-        # Returning value in addition to Entry elides the redundant `loads(dumps(...))` when obj_store is True.
+        if perf_logger.isEnabledFor(logging.DEBUG):
+            perf_logger.debug(
+                json.dumps(
+                    {
+                        "event": "serialize",
+                        "call_id": call_id,
+                        "duration": (stop - mid).total_seconds(),
+                    }
+                )
+            )
+            perf_logger.debug(
+                json.dumps(
+                    {
+                        "event": "inner_function",
+                        "call_id": call_id,
+                        "duration": (mid - start).total_seconds(),
+                    }
+                )
+            )
+
         return (
             # pyright doesn't know attrs __init__, hence type ignore
             Entry(  # type: ignore
@@ -505,33 +573,50 @@ class Memoized(Generic[FuncParams, FuncReturn]):
                 obj_store=self._use_obj_store,
             ),
             value,
+            # Returning value in addition to Entry elides the redundant `loads(dumps(...))` when obj_store is True.
         )
 
+    def __getfrozenstate__(self) -> Callable[..., Any]:
+        return self._func
+
     def __call__(
         self, *args: FuncParams.args, **kwargs: FuncParams.kwargs
     ) -> FuncReturn:
         call_start = datetime.datetime.now()
 
-        would_hit, key, obj_key = self._would_hit(*args, **kwargs)
+        call_id = random.randint(0, 2 ** 64 - 1)
+
+        would_hit, key, obj_key = self._would_hit(call_id, *args, **kwargs)
+
+        if ops_logger.isEnabledFor(logging.DEBUG):
+            ops_logger.debug(
+                json.dumps(
+                    {
+                        "event": "hit" if would_hit else "miss",
+                        "call_id": call_id,
+                        "name": self.name,
+                        "key": key,
+                        "obj_key": obj_key,
+                        # "args_kwargs": ellipsize(str(args) + " " + str(kwargs), 60),
+                    }
+                )
+            )
 
         if would_hit:
-            self.group.logger.getChild("hit").debug(
-                "Hit name=%r key=%s obj_key=0x%x args_kwargs=%r",
-                self.name,
-                key2str(key),
-                obj_key,
-                ellipsize(str(args) + " " + str(kwargs), 60),
-            )
 
             # Do the lookup
             with self.group._memory_lock:
                 entry = self.group._index[key]
-                value_ser = self.group._obj_store[obj_key] if entry.obj_store else None
+                with perf_ctx("obj_load", call_id=call_id):
+                    value_ser = (
+                        self.group._obj_store[obj_key] if entry.obj_store else None
+                    )
 
             # Deserialize
             if entry.obj_store:
                 assert value_ser is not None
-                value = cast(FuncReturn, self._pickler.loads(value_ser))
+                with perf_ctx("deserialize", call_id=call_id):
+                    value = cast(FuncReturn, self._pickler.loads(value_ser))
             else:
                 value = cast(FuncReturn, entry.value)
 
@@ -540,16 +625,9 @@ class Memoized(Generic[FuncParams, FuncReturn]):
                 self.group.time_saved[self.name] += entry.function_time
                 self.group._replacement_policy.access(key, entry)
         else:
-            self.group.logger.getChild("miss").debug(
-                "Miss name=%r key=%s obj_key=0x%x args_kwargs=%r",
-                self.name,
-                key2str(key),
-                obj_key,
-                ellipsize(str(args) + " " + str(kwargs), 60),
-            )
 
             # Do the recompute
-            entry, value = self._recompute(obj_key, *args, **kwargs)
+            entry, value = self._recompute(call_id, obj_key, *args, **kwargs)
 
             # Do the store
             with self.group._memory_lock:
@@ -581,13 +659,26 @@ class Memoized(Generic[FuncParams, FuncReturn]):
 
             tc = self.group.time_cost[self.name]
             ts = self.group.time_saved[self.name]
-            if ts < tc and tc.total_seconds() > 5:
-                warnings.warn(
-                    f"Caching {self._func} cost {tc.total_seconds():.1f}s but only saved {ts.total_seconds():.1f}s",
-                    UserWarning,
-                )
 
-            return value
+        if perf_logger.isEnabledFor(logging.DEBUG):
+            perf_logger.debug(
+                json.dumps(
+                    {
+                        "name": self.name,
+                        "event": "outer_function",
+                        "call_id": call_id,
+                        "hit": would_hit,
+                        "duration": (call_stop - call_start).total_seconds(),
+                    }
+                )
+            )
+        if ts < tc and tc.total_seconds() > 5:
+            warnings.warn(
+                f"Caching {self._func} cost {tc.total_seconds():.1f}s but only saved {ts.total_seconds():.1f}s",
+                UserWarning,
+            )
+
+        return value
 
     def _compress(self, obj: Any) -> Any:
         if self._lossy_compression:
@@ -595,26 +686,29 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         else:
             return obj
 
-    def would_hit(self, *args: FuncParams.args, **kwargs: FuncParams.kwargs) -> bool:
-        return self._would_hit(*args, **kwargs)[0]
+    def would_hit(
+        self, call_id: int, *args: FuncParams.args, **kwargs: FuncParams.kwargs
+    ) -> bool:
+        return self._would_hit(call_id, *args, **kwargs)[0]
 
     def _would_hit(
-        self, *args: FuncParams.args, **kwargs: FuncParams.kwargs
+        self, call_id: int, *args: FuncParams.args, **kwargs: FuncParams.kwargs
     ) -> Tuple[bool, Tuple[Any, ...], int]:
         # pylint: disable=protected-access
 
-        # Note that the system state and name or so small already, it isn't worth hashing them.
-        # They are also used by other Memoized functions in the same MemoizedGroup.
-        # We will only hash the potentially large key items that are used exclusively by this Memoized function.
-        key = (
-            # Group is a friend class, hence type ignore
-            hashable(self.group._system_state()),  # pylint: disable=protected-access
-            hashable(self.name),
-            self._compress(hashable(self._func_state())),
-            self._compress(hashable(self._args2key(*args, **kwargs))),
-            self._compress(hashable(self._args2ver(*args, **kwargs))),
-        )
-        obj_key = determ_hash(key)
+        with perf_ctx("hash", call_id=call_id):
+            # Note that the system state and name or so small already, it isn't worth hashing them.
+            # They are also used by other Memoized functions in the same MemoizedGroup.
+            # We will only hash the potentially large key items that are used exclusively by this Memoized function.
+            key = (
+                # Group is a friend class, hence type ignore
+                freeze(self.group._system_state()),  # pylint: disable=protected-access
+                freeze(self.name),
+                self._compress(freeze(self._func_state())),
+                self._compress(freeze(self._args2key(*args, **kwargs))),
+                self._compress(freeze(self._args2ver(*args, **kwargs))),
+            )
+            obj_key = determ_hash(key)
         with self.group._memory_lock:
             if self.group._fine_grain_persistence:
                 self.group._index_read()
